@@ -1,18 +1,18 @@
 /* A Scheme reader compiler for Guile.
 
-   Copyright (C) 2005, 2006, 2007, 2008  Ludovic Courts <ludo@gnu.org>
+   Copyright (C) 2005, 2006, 2007, 2008, 2009, 2010, 2012  Ludovic Courts <ludo@gnu.org>
 
    Part of the code here (a few `scm_token_reader_t' functions below) is
    based on Guile code released under the GNU LGPL (file `read.c') which
    contains the following copyright line:
 
-   Copyright (C) 1995,1996,1997,1999,2000,2001,2003, 2004 Free Software
-   Foundation, Inc.
+   Copyright (C) 1995, 1996, 1997, 1999, 2000, 2001, 2003, 2004, 2006,
+     2007, 2008, 2009, 2010, 2011, 2012 Free Software Foundation, Inc.
 
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
+   the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
@@ -47,53 +47,121 @@
 
 
 /* `isblank' is only in C99.  */
-#define CHAR_IS_BLANK_(_chr)					\
-  (((_chr) == ' ') || ((_chr) == '\t') || ((_chr) == '\n')	\
-   || ((_chr) == '\f') || ((_chr) == '\r'))
 
+#define BLANK_CHARS_ " \t\n\f\r"
 #ifdef MSDOS
-# define CHAR_IS_BLANK(_chr)			\
-  ((CHAR_IS_BLANK_ (chr)) || ((_chr) == 26))
+# define BLANK_CHARS  BLANK_CHARS_ "\032"
 #else
-# define CHAR_IS_BLANK CHAR_IS_BLANK_
+# define BLANK_CHARS  BLANK_CHARS_
 #endif
 
+#define CHAR_IS_BLANK(c)			\
+  (index (BLANK_CHARS, (c)) != NULL)
+
 
 /* R5RS one-character delimiters (see section 7.1.1, ``Lexical
    structure'').  */
-#define CHAR_IS_R5RS_DELIMITER(c)				\
-  (CHAR_IS_BLANK (c)						\
-   || (c == ')') || (c == '(') || (c == ';') || (c == '"'))
+#define R5RS_DELIMITERS  BLANK_CHARS "();\""
 
-#define CHAR_IS_R6RS_DELIMITER(c)				\
-  (CHAR_IS_R5RS_DELIMITER (c) || (c == '[') || (c == ']'))
+#define CHAR_IS_R5RS_DELIMITER(c)		\
+  (index (R5RS_DELIMITERS, (c)) != NULL)
+
+#define R6RS_DELIMITERS  R5RS_DELIMITERS "[]"
+
+#define CHAR_IS_R6RS_DELIMITER(c)		\
+  (index (R6RS_DELIMITERS, (c)) != NULL)
 
 
-/* Helper function similar to `scm_read_token ()'.  Read from PORT until a
-   whitespace is read.  */
-static inline void
-read_token (SCM port, char *buf, size_t buf_size, size_t *read)
+/* Read from PORT until a delimiter (e.g., a whitespace) is read.  Put the
+   result in the pre-allocated buffer BUF.  Return zero if the whole token has
+   fewer than BUF_SIZE bytes, non-zero otherwise. READ will be set the number of
+   bytes actually read.  */
+int
+read_token (SCM port, char *buf, size_t buf_size, const char *delimiters,
+	    size_t *read)
 {
   *read = 0;
 
-  while (buf_size)
+  while (*read < buf_size)
     {
       int chr;
-      chr = scm_getc (port);
+
+      chr = scm_get_byte_or_eof (port);
+
       if (chr == EOF)
-	break;
-      else if (CHAR_IS_R5RS_DELIMITER (chr))
+	return 0;
+      else if (index (delimiters, chr) != NULL)
 	{
-	  scm_ungetc (chr, port);
-	  break;
+	  scm_unget_byte (chr, port);
+	  return 0;
+	}
+      else
+	{
+	  *buf = (char) chr;
+	  buf++, (*read)++;
 	}
-
-      *(buf++) = (char)chr;
-      (*read)++;
-
-      if (*read == buf_size)
-	break;
     }
+
+  return 1;
+}
+
+/* Like `read_token', but return either BUFFER, or a GC-allocated buffer
+   if the token doesn't fit in BUFFER_SIZE bytes.  */
+static char *
+read_complete_token (SCM port, char *buffer, size_t buffer_size,
+		     const char *delimiters, size_t *read)
+{
+  int overflow = 0;
+  size_t bytes_read, overflow_size = 0;
+  char *overflow_buffer = NULL;
+
+  do
+    {
+      overflow = read_token (port, buffer, buffer_size,
+			     delimiters, &bytes_read);
+      if (bytes_read == 0)
+        break;
+      if (overflow || overflow_size != 0)
+        {
+          if (overflow_size == 0)
+            {
+#ifdef HAVE_SCM_GC_MALLOC_POINTERLESS
+              overflow_buffer = scm_gc_malloc_pointerless (bytes_read, "read");
+#else
+	      overflow_buffer = scm_malloc (bytes_read);
+#endif
+              memcpy (overflow_buffer, buffer, bytes_read);
+              overflow_size = bytes_read;
+            }
+          else
+            {
+	      char *new_buf =
+#ifdef HAVE_SCM_GC_MALLOC_POINTERLESS
+		scm_gc_malloc_pointerless (overflow_size + bytes_read, "read");
+#else
+	        scm_malloc (overflow_size + bytes_read);
+#endif
+
+	      memcpy (new_buf, overflow_buffer, overflow_size);
+              memcpy (new_buf + overflow_size, buffer, bytes_read);
+
+#ifdef HAVE_SCM_GC_MALLOC_POINTERLESS
+	      free (overflow_buffer);
+#endif
+
+	      overflow_buffer = new_buf;
+              overflow_size += bytes_read;
+            }
+        }
+    }
+  while (overflow);
+
+  if (overflow_size)
+    *read = overflow_size;
+  else
+    *read = bytes_read;
+
+  return (overflow_size > 0 ? overflow_buffer : buffer);
 }
 
 
@@ -106,7 +174,8 @@ read_token (SCM port, char *buf, size_t buf_size, size_t *read)
 static int
 _flush_ws (SCM port, const char *eoferr)
 {
-  register int c;
+  scm_t_wchar c;
+
   while (1)
     switch (c = scm_getc (port))
       {
@@ -149,7 +218,7 @@ _flush_ws (SCM port, const char *eoferr)
 #define scm_flush_ws _flush_ws
 
 SCM
-scm_read_sexp (int chr, SCM port, scm_reader_t reader,
+scm_read_sexp (scm_t_wchar chr, SCM port, scm_reader_t reader,
 	       scm_reader_t top_level_reader)
 #define FUNC_NAME "scm_read_sexp"
 {
@@ -275,8 +344,46 @@ exit:
 
 #undef scm_flush_ws
 
+/* Number of 32-bit codepoints in the buffer used to read strings.  */
+#define READER_STRING_BUFFER_SIZE 128
+
+/* Read a hexadecimal number NDIGITS in length.  Put its value into the variable
+   C.  If TERMINATOR is non-null, terminate early if the TERMINATOR character is
+   found.  */
+#define SCM_READ_HEX_ESCAPE(ndigits, terminator)                   \
+  do                                                               \
+    {                                                              \
+      scm_t_wchar a;                                               \
+      size_t i = 0;                                                \
+      c = 0;                                                       \
+      while (i < ndigits)                                          \
+        {                                                          \
+          a = scm_getc (port);                                     \
+          if (a == EOF)                                            \
+            goto str_eof;                                          \
+          if (terminator                                           \
+              && (a == (scm_t_wchar) terminator)                   \
+              && (i > 0))                                          \
+            break;                                                 \
+          if ('0' <= a && a <= '9')                                \
+            a -= '0';                                              \
+          else if ('A' <= a && a <= 'F')                           \
+            a = a - 'A' + 10;                                      \
+          else if ('a' <= a && a <= 'f')                           \
+            a = a - 'a' + 10;                                      \
+          else                                                     \
+            {                                                      \
+              c = a;                                               \
+              goto bad_escaped;                                    \
+            }                                                      \
+          c = c * 16 + a;                                          \
+          i ++;                                                    \
+        }                                                          \
+    } while (0)
+
+/* FIXME: Add a version that supports R6RS escapes.  */
 SCM
-scm_read_string (int chr, SCM port, scm_reader_t scm_reader,
+scm_read_string (scm_t_wchar chr, SCM port, scm_reader_t scm_reader,
 		 scm_reader_t top_level_reader)
 #define FUNC_NAME "scm_read_string"
 {
@@ -284,94 +391,90 @@ scm_read_string (int chr, SCM port, scm_reader_t scm_reader,
      object (the string returned).  */
 
   SCM str = SCM_EOL;
-  char c_str[512];
-  unsigned c_str_len = 0;
-  int c;
+  size_t c_str_len = 0;
+  scm_t_wchar c, c_str[READER_STRING_BUFFER_SIZE];
 
   while ('"' != (c = scm_getc (port)))
     {
       if (c == EOF)
-	str_eof: scm_i_input_error (FUNC_NAME, port,
-				    "end of file in string constant",
-				    SCM_EOL);
+        {
+        str_eof:
+          scm_i_input_error (FUNC_NAME, port,
+                             "end of file in string constant", SCM_EOL);
+        }
 
-      if (c_str_len + 1 >= sizeof (c_str))
+      if (c_str_len + 1 >= READER_STRING_BUFFER_SIZE)
 	{
-	  /* Flush the C buffer onto a Scheme string.  */
-	  str = scm_cons (scm_from_locale_stringn (c_str, c_str_len), str);
-
+	  str = scm_cons (scm_from_utf32_stringn (c_str, c_str_len), str);
 	  c_str_len = 0;
 	}
 
       if (c == '\\')
-	switch (c = scm_getc (port))
-	  {
-	  case EOF:
-	    goto str_eof;
-	  case '"':
-	  case '\\':
-	    break;
-#if 0 /* SCM_ENABLE_ELISP */
-	  case '(':
-	  case ')':
-	    if (SCM_ESCAPED_PARENS_P)
+        {
+          switch (c = scm_getc (port))
+            {
+            case EOF:
+              goto str_eof;
+            case '"':
+            case '\\':
+              break;
+            case '\n':
+              continue;
+            case '0':
+              c = '\0';
+              break;
+            case 'f':
+              c = '\f';
+              break;
+            case 'n':
+              c = '\n';
+              break;
+            case 'r':
+              c = '\r';
+              break;
+            case 't':
+              c = '\t';
+              break;
+            case 'a':
+              c = '\007';
+              break;
+            case 'v':
+              c = '\v';
+              break;
+            case 'b':
+              c = '\010';
+              break;
+            case 'x':
+	      SCM_READ_HEX_ESCAPE (2, '\0');
+              break;
+            case 'u':
+	      SCM_READ_HEX_ESCAPE (4, '\0');
 	      break;
-	    goto bad_escaped;
-#endif
-	  case '\n':
-	    continue;
-	  case '0':
-	    c = '\0';
-	    break;
-	  case 'f':
-	    c = '\f';
-	    break;
-	  case 'n':
-	    c = '\n';
-	    break;
-	  case 'r':
-	    c = '\r';
-	    break;
-	  case 't':
-	    c = '\t';
-	    break;
-	  case 'a':
-	    c = '\007';
-	    break;
-	  case 'v':
-	    c = '\v';
-	    break;
-	  case 'x':
-	    {
-	      int a, b;
-	      a = scm_getc (port);
-	      if (a == EOF) goto str_eof;
-	      b = scm_getc (port);
-	      if (b == EOF) goto str_eof;
-	      if      ('0' <= a && a <= '9') a -= '0';
-	      else if ('A' <= a && a <= 'F') a = a - 'A' + 10;
-	      else if ('a' <= a && a <= 'f') a = a - 'a' + 10;
-	      else goto bad_escaped;
-	      if      ('0' <= b && b <= '9') b -= '0';
-	      else if ('A' <= b && b <= 'F') b = b - 'A' + 10;
-	      else if ('a' <= b && b <= 'f') b = b - 'a' + 10;
-	      else goto bad_escaped;
-	      c = a * 16 + b;
+            case 'U':
+	      SCM_READ_HEX_ESCAPE (6, '\0');
 	      break;
-	    }
-	  default:
-	  bad_escaped:
-	    scm_i_input_error(FUNC_NAME, port,
-			      "illegal character in escape sequence: ~S",
-			      scm_list_1 (SCM_MAKE_CHAR (c)));
-	  }
+
+            default:
+	    bad_escaped:
+              scm_i_input_error (FUNC_NAME, port,
+                                 "invalid character in escape sequence: ~S",
+                                 scm_list_1 (SCM_MAKE_CHAR (c)));
+            }
+        }
+
       c_str[c_str_len++] = c;
     }
 
-  if (c_str_len > 0)
-    str = scm_cons (scm_from_locale_stringn (c_str, c_str_len), str);
+  if (scm_is_null (str))
+    /* Fast path: we got a string that fits in C_STR.  */
+    str = scm_from_utf32_stringn (c_str, c_str_len);
+  else
+    {
+      if (c_str_len > 0)
+	str = scm_cons (scm_from_utf32_stringn (c_str, c_str_len), str);
 
-  str = scm_string_concatenate (scm_reverse_x (str, SCM_EOL));
+      str = scm_string_concatenate_reverse (str, SCM_UNDEFINED, SCM_UNDEFINED);
+    }
 
   return str;
 }
@@ -393,12 +496,12 @@ scm_read_string (int chr, SCM port, scm_reader_t scm_reader,
 
 #define SYMBOL_TR_NAME scm_read_guile_mixed_case_symbol
 #define NUMBER_TR_NAME scm_read_guile_number
-#define SYMBOL_TR_TRANSFORM_CHARACTER(_c)  do {} while (0)
-#define CHAR_IS_DELIMITER(_c)   (CHAR_IS_R5RS_DELIMITER (_c))
+#define SYMBOL_TR_TRANSFORM_CHARACTERS(s)  do {} while (0)
+#define DELIMITERS     R5RS_DELIMITERS
 
 #include "symbol-token-reader.c"
 
-#undef SYMBOL_TR_TRANSFORM_CHARACTER
+#undef SYMBOL_TR_TRANSFORM_CHARACTERS
 #undef NUMBER_TR_NAME
 #undef SYMBOL_TR_NAME
 
@@ -406,58 +509,57 @@ scm_read_string (int chr, SCM port, scm_reader_t scm_reader,
 
 #define SYMBOL_TR_NAME scm_read_r5rs_lower_case_symbol
 #define NUMBER_TR_NAME scm_read_r5rs_lower_case_number
-#define SYMBOL_TR_TRANSFORM_CHARACTER(_c)	\
-  do { (_c) = tolower (_c); } while (0)
+#define SYMBOL_TR_TRANSFORM_CHARACTERS(s)		\
+  do { s = scm_string_downcase_x (s); } while (0)
 
 #include "symbol-token-reader.c"
 
-#undef SYMBOL_TR_TRANSFORM_CHARACTER
+#undef SYMBOL_TR_TRANSFORM_CHARACTERS
 #undef NUMBER_TR_NAME
 #undef SYMBOL_TR_NAME
 
 #define SYMBOL_TR_NAME scm_read_r5rs_upper_case_symbol
 #define NUMBER_TR_NAME scm_read_r5rs_upper_case_number
-#define SYMBOL_TR_TRANSFORM_CHARACTER(_c)	\
-  do { (_c) = toupper (_c); } while (0)
+#define SYMBOL_TR_TRANSFORM_CHARACTERS(s)		\
+  do { s = scm_string_upcase_x (s); } while (0)
 
 #include "symbol-token-reader.c"
 
-#undef SYMBOL_TR_TRANSFORM_CHARACTER
+#undef SYMBOL_TR_TRANSFORM_CHARACTERS
 #undef NUMBER_TR_NAME
 #undef SYMBOL_TR_NAME
-#undef CHAR_IS_DELIMITER
+#undef DELIMITERS
 
 /* FIXME: Check whether this really corresponds to R6RS.  */
 
 #define SYMBOL_TR_NAME scm_read_r6rs_symbol
 #define NUMBER_TR_NAME scm_read_r6rs_number
-#define SYMBOL_TR_TRANSFORM_CHARACTER(_c)  do {} while (0)
-#define CHAR_IS_DELIMITER(_c)   (CHAR_IS_R6RS_DELIMITER (_c))
+#define SYMBOL_TR_TRANSFORM_CHARACTERS(s)  do {} while (0)
+#define DELIMITERS     R6RS_DELIMITERS
 #include "symbol-token-reader.c"
 
-#undef SYMBOL_TR_TRANSFORM_CHARACTER
+#undef SYMBOL_TR_TRANSFORM_CHARACTERS
 #undef NUMBER_TR_NAME
 #undef SYMBOL_TR_NAME
-#undef CHAR_IS_DELIMITER
+#undef DELIMITERS
 
 #define SYMBOL_TR_NAME scm_read_brace_free_symbol
 #define NUMBER_TR_NAME scm_read_brace_free_number
-#define SYMBOL_TR_TRANSFORM_CHARACTER(_c)  do {} while (0)
-#define CHAR_IS_DELIMITER(_c)			\
-  ((CHAR_IS_R5RS_DELIMITER (_c))		\
-   || ((_c) == '}') || ((_c) == '{'))
+#define SYMBOL_TR_TRANSFORM_CHARACTERS(s)  do {} while (0)
+#define DELIMITERS     R5RS_DELIMITERS "{}"
 
 #include "symbol-token-reader.c"
 
-#undef SYMBOL_TR_TRANSFORM_CHARACTER
+#undef SYMBOL_TR_TRANSFORM_CHARACTERS
 #undef NUMBER_TR_NAME
 #undef SYMBOL_TR_NAME
+#undef DELIMITERS
 #undef SYMBOL_TR_VALID_CHARACTER
 
 
 
 SCM
-scm_read_number_and_radix (int chr, SCM port, scm_reader_t scm_reader,
+scm_read_number_and_radix (scm_t_wchar chr, SCM port, scm_reader_t scm_reader,
 			   scm_reader_t top_level_reader)
 {
   int c;
@@ -539,7 +641,7 @@ scm_read_number_and_radix (int chr, SCM port, scm_reader_t scm_reader,
 }
 
 SCM
-scm_read_quote (int chr, SCM port, scm_reader_t scm_reader,
+scm_read_quote (scm_t_wchar chr, SCM port, scm_reader_t scm_reader,
 		scm_reader_t top_level_reader)
 {
   SCM p;
@@ -556,7 +658,7 @@ scm_read_quote (int chr, SCM port, scm_reader_t scm_reader,
 
     case ',':
       {
-	int c;
+	scm_t_wchar c;
 
 	c = scm_getc (port);
 	if ('@' == c)
@@ -581,11 +683,62 @@ scm_read_quote (int chr, SCM port, scm_reader_t scm_reader,
   return p;
 }
 
+SCM_SYMBOL (sym_syntax, "syntax");
+SCM_SYMBOL (sym_quasisyntax, "quasisyntax");
+SCM_SYMBOL (sym_unsyntax, "unsyntax");
+SCM_SYMBOL (sym_unsyntax_splicing, "unsyntax-splicing");
+
+/* Read R6RS-style syntax quotes: #', #`, #,.  Adapted from Guile's `read.c'
+   where it appeared in commit 34f3d47d, dated 2009-05-28.  */
 SCM
-scm_read_semicolon_comment (int chr, SCM port, scm_reader_t scm_reader,
+scm_read_r6rs_syntax_quote (scm_t_wchar chr, SCM port, scm_reader_t scm_reader,
 			    scm_reader_t top_level_reader)
 {
-  int c;
+  SCM p;
+
+  switch (chr)
+    {
+    case '`':
+      p = sym_quasisyntax;
+      break;
+
+    case '\'':
+      p = sym_syntax;
+      break;
+
+    case ',':
+      {
+       scm_t_wchar c;
+
+       c = scm_getc (port);
+       if ('@' == c)
+         p = sym_unsyntax_splicing;
+       else
+         {
+           scm_ungetc (c, port);
+           p = sym_unsyntax;
+         }
+       break;
+      }
+
+    default:
+      fprintf (stderr, "%s: unhandled syntax character (%i)\n",
+              "scm_read_syntax", chr);
+      abort ();
+    }
+
+  p = scm_cons2 (p, scm_call_reader (top_level_reader, port, 0,
+				     top_level_reader),
+		 SCM_EOL);
+
+  return p;
+}
+
+SCM
+scm_read_semicolon_comment (scm_t_wchar chr, SCM port, scm_reader_t scm_reader,
+			    scm_reader_t top_level_reader)
+{
+  scm_t_wchar c;
 
   for (c = scm_getc (port);
        (c != EOF) && (c != '\n');
@@ -598,7 +751,7 @@ scm_read_semicolon_comment (int chr, SCM port, scm_reader_t scm_reader,
 /* Sharp readers, i.e. readers called after a `#' sign has been read.  */
 
 SCM
-scm_read_boolean (int chr, SCM port, scm_reader_t scm_reader,
+scm_read_boolean (scm_t_wchar chr, SCM port, scm_reader_t scm_reader,
 		  scm_reader_t top_level_reader)
 {
   switch (chr)
@@ -615,15 +768,43 @@ scm_read_boolean (int chr, SCM port, scm_reader_t scm_reader,
   return SCM_UNSPECIFIED;
 }
 
+#ifndef HAVE_SCM_CHARNAMES
+
+static const char *const scm_charnames[] =
+{
+  "nul","soh","stx","etx","eot","enq","ack","bel",
+   "bs", "ht", "newline", "vt", "np", "cr", "so", "si",
+  "dle","dc1","dc2","dc3","dc4","nak","syn","etb",
+  "can", "em","sub","esc", "fs", "gs", "rs", "us",
+  "space", "sp", "nl", "tab", "backspace", "return", "page", "null", "del"
+};
+
+#define scm_n_charnames  (sizeof (scm_charnames) / sizeof (char *))
+
+#endif /* !HAVE_SCM_CHARNAMES */
+
+#ifndef HAVE_SCM_CHARNUMS
+
+static const char scm_charnums[] =
+"\000\001\002\003\004\005\006\007\
+\010\011\012\013\014\015\016\017\
+\020\021\022\023\024\025\026\027\
+\030\031\032\033\034\035\036\037\
+  \n\t\b\r\f\0\177";
+
+#endif /* !HAVE_SCM_CHARNUMS */
+
+
 SCM
-scm_read_character (int chr, SCM port, scm_reader_t reader,
+scm_read_character (scm_t_wchar chr, SCM port, scm_reader_t reader,
 		    scm_reader_t top_level_reader)
 {
   unsigned c;
   char charname[100];
   size_t charname_len;
 
-  read_token (port, charname, sizeof (charname), &charname_len);
+  read_token (port, charname, sizeof charname, R5RS_DELIMITERS,
+	      &charname_len);
 
   if (charname_len == 0)
     {
@@ -664,7 +845,7 @@ scm_read_character (int chr, SCM port, scm_reader_t reader,
 }
 
 SCM
-scm_read_keyword (int chr, SCM port, scm_reader_t reader,
+scm_read_keyword (scm_t_wchar chr, SCM port, scm_reader_t reader,
 		  scm_reader_t top_level_reader)
 {
   SCM symbol;
@@ -686,7 +867,7 @@ scm_read_keyword (int chr, SCM port, scm_reader_t reader,
 }
 
 SCM
-scm_read_vector (int chr, SCM port, scm_reader_t reader,
+scm_read_vector (scm_t_wchar chr, SCM port, scm_reader_t reader,
 		 scm_reader_t top_level_reader)
 {
   /* Note: We call `scm_read_sexp ()' rather than READER here in order to
@@ -698,14 +879,14 @@ scm_read_vector (int chr, SCM port, scm_reader_t reader,
 }
 
 SCM
-scm_read_srfi4_vector (int chr, SCM port, scm_reader_t scm_reader,
+scm_read_srfi4_vector (scm_t_wchar chr, SCM port, scm_reader_t scm_reader,
 		       scm_reader_t top_level_reader)
 {
   return scm_i_read_array (port, chr);
 }
 
 SCM
-scm_read_guile_bit_vector (int chr, SCM port, scm_reader_t scm_reader,
+scm_read_guile_bit_vector (scm_t_wchar chr, SCM port, scm_reader_t scm_reader,
 			   scm_reader_t top_level_reader)
 {
   /* Read the `#*10101'-style read syntax for bit vectors in Guile.  This is
@@ -726,14 +907,14 @@ scm_read_guile_bit_vector (int chr, SCM port, scm_reader_t scm_reader,
 }
 
 SCM
-scm_read_scsh_block_comment (int chr, SCM port, scm_reader_t reader,
+scm_read_scsh_block_comment (scm_t_wchar chr, SCM port, scm_reader_t reader,
 			     scm_reader_t top_level_reader)
 {
   int bang_seen = 0;
 
   for (;;)
     {
-      int c = scm_getc (port);
+      scm_t_wchar c = scm_getc (port);
 
       if (c == EOF)
 	scm_i_input_error (__FUNCTION__, port,
@@ -751,7 +932,7 @@ scm_read_scsh_block_comment (int chr, SCM port, scm_reader_t reader,
 }
 
 SCM
-scm_read_srfi30_block_comment (int chr, SCM port, scm_reader_t reader,
+scm_read_srfi30_block_comment (scm_t_wchar chr, SCM port, scm_reader_t reader,
 			       scm_reader_t top_level_reader)
 {
   /* Unlike SCSH-style block comments, SRFI-30 block comments may be nested.
@@ -761,7 +942,7 @@ scm_read_srfi30_block_comment (int chr, SCM port, scm_reader_t reader,
 
   while (nesting_level > 0)
     {
-      int c = scm_getc (port);
+      scm_t_wchar c = scm_getc (port);
 
       if (c == EOF)
 	scm_i_input_error (__FUNCTION__, port,
@@ -791,7 +972,7 @@ scm_read_srfi30_block_comment (int chr, SCM port, scm_reader_t reader,
 }
 
 SCM
-scm_read_srfi62_sexp_comment (int chr, SCM port, scm_reader_t reader,
+scm_read_srfi62_sexp_comment (scm_t_wchar chr, SCM port, scm_reader_t reader,
 			      scm_reader_t top_level_reader)
 {
   /* Skip the S-expression that follows the semi-colon.  */
@@ -801,7 +982,7 @@ scm_read_srfi62_sexp_comment (int chr, SCM port, scm_reader_t reader,
 }
 
 SCM
-scm_read_extended_symbol (int chr, SCM port, scm_reader_t scm_reader,
+scm_read_extended_symbol (scm_t_wchar chr, SCM port, scm_reader_t scm_reader,
 			  scm_reader_t top_level_reader)
 {
   /* Guile's extended symbol read syntax looks like this:
@@ -859,35 +1040,31 @@ scm_read_extended_symbol (int chr, SCM port, scm_reader_t scm_reader,
 
 /* Skribe's read syntax.  So-called sk-exps.  */
 SCM
-scm_read_skribe_exp (int chr, SCM port, scm_reader_t reader,
+scm_read_skribe_exp (scm_t_wchar chr, SCM port, scm_reader_t reader,
 		     scm_reader_t top_level_reader)
 {
-  int c, escaped = 0;
-  char c_literal[2048];
+  scm_t_wchar c, escaped = 0;
+  scm_t_wchar c_literal[1024];
   size_t c_literal_len = 0;
-  SCM result = SCM_EOL, s_literal;
+  SCM result = SCM_EOL, s_literal = SCM_EOL;
 
 #define FLUSH_STRING()							\
   do									\
     {									\
       s_literal =							\
-	scm_string_append (scm_list_2					\
-			   (s_literal,					\
-			    scm_from_locale_stringn (c_literal,		\
-						     c_literal_len)));	\
+	scm_cons (scm_from_utf32_stringn (c_literal, c_literal_len),	\
+		  s_literal);						\
       c_literal_len = 0;						\
     }									\
   while (0)
 
-  s_literal = scm_c_make_string (0, SCM_MAKE_CHAR ('X'));
-
   for (c = scm_getc (port);
        (c != EOF) && ((c != ']') || (escaped));
        c = scm_getc (port))
     {
       if (escaped)
 	{
-	  c_literal[c_literal_len++] = (char)c;
+	  c_literal[c_literal_len++] = c;
 	  escaped = 0;
 	}
       else
@@ -901,8 +1078,11 @@ scm_read_skribe_exp (int chr, SCM port, scm_reader_t reader,
 		  SCM subexp;
 
 		  FLUSH_STRING ();
+		  s_literal = scm_string_concatenate_reverse (s_literal,
+							      SCM_UNDEFINED,
+							      SCM_UNDEFINED);
 		  result = scm_cons (s_literal, result);
-		  s_literal = scm_c_make_string (0, SCM_MAKE_CHAR ('X'));
+		  s_literal = SCM_EOL;
 
 		  scm_ungetc (c, port);
 		  subexp = scm_cons2 (scm_sym_unquote,
@@ -916,7 +1096,7 @@ scm_read_skribe_exp (int chr, SCM port, scm_reader_t reader,
 		{
 		  c_literal[c_literal_len++] = ',';
 		  if (c != EOF)
-		    c_literal[c_literal_len++] = (char)c;
+		    c_literal[c_literal_len++] = c;
 		}
 	      break;
 
@@ -925,7 +1105,7 @@ scm_read_skribe_exp (int chr, SCM port, scm_reader_t reader,
 	      break;
 
 	    default:
-	      c_literal[c_literal_len++] = (char)c;
+	      c_literal[c_literal_len++] = c;
 	    }
 	}
 
@@ -936,8 +1116,12 @@ scm_read_skribe_exp (int chr, SCM port, scm_reader_t reader,
     }
 
   FLUSH_STRING ();
-  if (scm_c_string_length (s_literal) > 0)
-    result = scm_cons (s_literal, result);
+  if (scm_is_pair (s_literal))
+    {
+      s_literal = scm_string_concatenate_reverse (s_literal, SCM_UNDEFINED,
+						  SCM_UNDEFINED);
+      result = scm_cons (s_literal, result);
+    }
 
   result = scm_reverse_x (result, SCM_EOL);
   result = scm_cons2 (scm_sym_quasiquote, result, SCM_EOL);
